<?php

namespace Wikibase\Repo\Store\Sql;

use MediaWiki\MediaWikiServices;
use Onoi\MessageReporter\MessageReporter;
use Onoi\MessageReporter\NullMessageReporter;
use Wikibase\DataModel\Entity\EntityId;
use Wikibase\Lib\Store\EntityRevisionLookup;
use Wikibase\Lib\Store\Sql\TermSqlIndex;
use Wikibase\Lib\TermIndexEntry;
use Wikimedia\Assert\Assert;
use Wikimedia\Rdbms\IDatabase;
use Wikimedia\Rdbms\ILBFactory;
use Wikimedia\Rdbms\ILoadBalancer;

/**
 * (Re)builds term index in the SQL table.
 * This can add missing information to the SQL table like missing full entity ID. It also removes
 * possible duplicate terms.
 * It can also ensure that all expected entity terms are stored in the term index, i.e. add
 * all possible missing terms of the given entity, and remove all possible no longer valid
 * terms of the entity, even if there is no other need for rebuilding the index
 * (i.e. all ID fields are populated, there are no duplicate entries).
 *
 * @license GPL-2.0-or-later
 * @author Katie Filbert < aude.wiki@gmail.com >
 */
class TermSqlIndexBuilder {

	const TABLE_NAME = 'wb_terms';

	/**
	 * @var ILBFactory
	 */
	private $loadBalancerFactory;

	/**
	 * @var TermSqlIndex
	 */
	private $termSqlIndex;

	/**
	 * @var SqlEntityIdPagerFactory
	 */
	private $entityIdPagerFactory;

	/**
	 * @var EntityRevisionLookup
	 */
	private $entityRevisionLookup;

	/**
	 * @var string[]
	 */
	private $entityTypes;

	/**
	 * @var int
	 */
	private $sleep;

	/**
	 * @var MessageReporter
	 */
	private $progressReporter;

	/**
	 * @var MessageReporter
	 */
	private $errorReporter;

	/**
	 * @var int
	 */
	private $batchSize = 1000;

	/**
	 * @var bool
	 */
	private $removeDuplicateTerms = false;

	/**
	 * @var bool
	 */
	private $rebuildAllEntityTerms = false;

	/**
	 * @var int|null
	 */
	private $fromId = null;

	/**
	 * @param ILBFactory $loadBalancerFactory
	 * @param TermSqlIndex $termSqlIndex
	 * @param SqlEntityIdPagerFactory $entityIdPagerFactory
	 * @param EntityRevisionLookup $entityRevisionLookup
	 * @param string[] $entityTypes
	 * @param int $sleep Sleep time between each batch
	 */
	public function __construct(
		ILBFactory $loadBalancerFactory,
		TermSqlIndex $termSqlIndex,
		SqlEntityIdPagerFactory $entityIdPagerFactory,
		EntityRevisionLookup $entityRevisionLookup,
		array $entityTypes,
		$sleep = 0
	) {
		$this->loadBalancerFactory = $loadBalancerFactory;
		$this->termSqlIndex = $termSqlIndex;
		$this->entityIdPagerFactory = $entityIdPagerFactory;
		$this->entityRevisionLookup = $entityRevisionLookup;
		$this->entityTypes = $entityTypes;
		$this->sleep = $sleep;

		$this->progressReporter = new NullMessageReporter();
		$this->errorReporter = new NullMessageReporter();
	}

	public function setProgressReporter( MessageReporter $reporter ) {
		$this->progressReporter = $reporter;
	}

	public function setErrorReporter( MessageReporter $reporter ) {
		$this->errorReporter = $reporter;
	}

	/**
	 * @param int $size
	 */
	public function setBatchSize( $size ) {
		$this->batchSize = $size;
	}

	public function rebuild() {
		foreach ( $this->entityTypes as $entityType ) {
			$this->rebuildForEntityType( $entityType );
		}
	}

	/**
	 * @param int $fromId
	 */
	public function setFromId( $fromId ) {
		Assert::parameterType( 'integer', $fromId, 'fromId' );

		$this->fromId = $fromId;
	}

	/**
	 * Enables or disables the complete rebuilding of entity terms, i.e. removing all existing
	 * terms of the entity, and adding its terms to the index again.
	 *
	 * @param bool $rebuildAllEntityTerms
	 */
	public function setRebuildAllEntityTerms( $rebuildAllEntityTerms ) {
		$this->rebuildAllEntityTerms = $rebuildAllEntityTerms;
	}

	/**
	 * Enables or disables checking if the index contains duplicate terms for the particular
	 * entity, and removing duplicates if found any.
	 *
	 * Note this setting is redundant if the complete rebuild of the entity term index is enabled.
	 *
	 * @param bool $removeDuplicateTerms
	 */
	public function setRemoveDuplicateTerms( $removeDuplicateTerms ) {
		$this->removeDuplicateTerms = $removeDuplicateTerms;
	}

	/**
	 * @param string $entityType
	 */
	private function rebuildForEntityType( $entityType ) {
		$idPager = $this->entityIdPagerFactory->newSqlEntityIdPager( [ $entityType ] );
		$lastIdProcessed = null;

		if ( $this->fromId !== null ) {
			$idPager->setPosition( $this->fromId );
		}

		$ticket = $this->loadBalancerFactory->getEmptyTransactionTicket( __METHOD__ );
		$loadBalancer = $this->loadBalancerFactory->getMainLB();

		while ( true ) {
			$entityIds = $idPager->fetchIds( $this->batchSize );

			if ( !$entityIds ) {
				break;
			}

			$this->rebuildTermsForBatch( $entityIds, $loadBalancer, $ticket );
			$this->progressReporter->reportMessage( 'Processed up to page '
				. $idPager->getPosition() . ' (' . end( $entityIds ) . ')' );

			if ( $this->sleep > 0 ) {
				sleep( $this->sleep );
			}
		}

		$this->progressReporter->reportMessage( "Done rebuilding $entityType terms" );
	}

	/**
	 * @param EntityId[] $entityIds
	 * @param ILoadBalancer $loadBalancer
	 * @param mixed $transactionTicket
	 */
	private function rebuildTermsForBatch( array $entityIds, ILoadBalancer $loadBalancer, $transactionTicket ) {
		$dbr = $loadBalancer->getConnection( DB_REPLICA );
		$dbw = $loadBalancer->getConnection( DB_MASTER );

		foreach ( $entityIds as $entityId ) {
			$this->rebuildEntityTerms( $dbr, $dbw, $entityId );
		}

		$loadBalancer->reuseConnection( $dbw );
		$loadBalancer->reuseConnection( $dbr );

		$this->loadBalancerFactory->commitAndWaitForReplication( __METHOD__, $transactionTicket );
	}

	private function rebuildEntityTerms( IDatabase $dbr, IDatabase $dbw, EntityId $entityId ) {
		if ( $this->rebuildAllEntityTerms ) {
			$this->rebuildAllTermsOfEntity( $entityId );
			return;
		}

		if ( $this->removeDuplicateTerms ) {
			$existingTerms = $this->termSqlIndex->getTermsOfEntity( $entityId );
			$duplicateTerms = $this->getDuplicateTerms( $existingTerms );
			if ( $duplicateTerms ) {
				$this->removeDuplicateTermsOfEntity( $dbw, $entityId, $duplicateTerms );
			}
		}
	}

	private function rebuildAllTermsOfEntity( EntityId $entityId ) {
		$serializedId = $entityId->getSerialization();

		$success = $this->termSqlIndex->deleteTermsOfEntity( $entityId );

		if ( !$success ) {
			$this->loadBalancerFactory->rollbackMasterChanges( __METHOD__ );
			$this->errorReporter->reportMessage(
				"Failed to delete terms of entity: $serializedId"
			);

			return;
		}

		$entityRevision = $this->entityRevisionLookup->getEntityRevision( $entityId );
		$success = $this->termSqlIndex->saveTermsOfEntity( $entityRevision->getEntity() );

		if ( !$success ) {
			$this->loadBalancerFactory->rollbackMasterChanges( __METHOD__ );
			$this->errorReporter->reportMessage(
				"Failed to save terms of entity: $serializedId"
			);
		}
	}

	/**
	 * @param IDatabase $db
	 * @param EntityId $entityId
	 * @param TermIndexEntry[] $duplicateTerms
	 */
	private function removeDuplicateTermsOfEntity( IDatabase $db, EntityId $entityId, array $duplicateTerms ) {
		$idConds = [ 'term_full_entity_id' => $entityId->getSerialization() ];

		/** @var TermIndexEntry $term */
		foreach ( $duplicateTerms as $term ) {
			MediaWikiServices::getInstance()->getStatsdDataFactory()->increment(
				'wikibase.repo.wb_terms.selectFieldValues.TermSqlIndexBuilder_removeDuplicateTermsOfEntity'
			);
			$rowIds = $db->selectFieldValues(
				self::TABLE_NAME,
				'term_row_id',
				array_merge(
					$idConds,
					[
						'term_language' => $term->getLanguage(),
						'term_type' => $term->getTermType(),
						'term_text' => $term->getText(),
					]
				),
				__METHOD__
			);

			if ( !$rowIds ) {
				continue;
			}

			array_shift( $rowIds );
			MediaWikiServices::getInstance()->getStatsdDataFactory()->updateCount(
				'wikibase.repo.wb_terms.delete.TermSqlIndexBuilder_removeDuplicateTermsOfEntity',
				count( $rowIds )
			);
			foreach ( $rowIds as $id ) {
				$db->delete( self::TABLE_NAME, [ 'term_row_id' => $id ], __METHOD__ );
			}
		}
	}

	/**
	 * @param TermIndexEntry[] $terms
	 *
	 * @return TermIndexEntry[]
	 */
	private function getDuplicateTerms( array $terms ) {
		$duplicateTerms = [];

		foreach ( $terms as $index => $term ) {
			foreach ( $terms as $otherIndex => $otherTerm ) {
				if ( $index === $otherIndex ) {
					continue;
				}

				if ( TermIndexEntry::compare( $term, $otherTerm ) === 0 ) {
					$key = implode(
						':',
						[ $term->getLanguage(), $term->getTermType(), $term->getText() ]
					);

					$duplicateTerms[$key] = $term;
				}
			}
		}

		return array_values( $duplicateTerms );
	}

}
